#define _CRT_SECURE_NO_WARNINGS

#define TMP_GRAPH_IO

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>
#include <algorithm>
#include "tmp_graph.h"

#define GRAPH_BASE_REAL int

#define GRAPH_NAMESPACE GOMORY_HU_TEST
#include "bk.inc"
#undef GRAPH_NAMESPACE

void ComputeCut(TmpGraph<int>* T, int _e, int* _x)
{
	int* p = new int[T->n];
	int i, e, dir;
	for (i=0; i<T->n; i++) p[i] = -1;
	for (e=0; e<T->m; e++)
	{
		if (e == _e) continue;
		int r = T->edges[2*e];
		while (p[r] >= 0) r = p[r]; 
		for (dir=0; dir<2; dir++)
		{
			i = T->edges[2*e+dir];
			while (i != r && i >= 0) { int j = p[i]; p[i] = r; i = j; }
		}
	}
	int s = -1, t = -1;
	for (i=0; i<T->n; i++)
	{
		if (p[i] < 0) { s = i; break; }
	}
	for (i++ ; i<T->n; i++)
	{
		if (p[i] < 0) { t = i; break; }
	}
	if (s < 0 || t < 0) { printf("error #1\n"); exit(1); }
	for (i++ ; i<T->n; i++)
	{
		if (p[i] < 0) { printf("error #2\n"); exit(1); }
	}
	for (i=0; i<T->n; i++)
	{
		int j = i;
		while (j != s && j != t) j = p[j];
		if (j == s) _x[i] = 0; else _x[i] = 1;
	}
	delete [] p;
}


int ComputeCut(TmpGraph<int>* input, int s, int t)
{
	int i, e, n = input->n;

	Buffer buf(1024);
	ReusableBuffer rbuf_common, rbuf[MAXFLOW_RBUF_NUM];
	typedef GOMORY_HU_TEST::Graph G;
	G g(n, input->m, &buf, &rbuf_common, rbuf);
	int* W = new int[n];
	memset(W, 0, n*sizeof(int));
	for (e=0; e<input->m; e++)
	{
		int i = input->edges[2*e], j = input->edges[2*e+1], w = input->weights[e];
		g.add_edge(i, j, w, w);
		W[i] += w;
		W[j] += w;
	}
	int W_max = 0;
	for (i=0; i<n; i++) { if (W_max < W[i]) W_max = W[i]; }
	W_max += 1;
	delete [] W;

	g.add_tweights(s, W_max, 0);
	g.add_tweights(t, 0, W_max);
	g.maxflow();

	int cost = 0;
	for (e=0; e<input->m; e++)
	{
		if (g.what_segment(input->edges[2*e]) != g.what_segment(input->edges[2*e+1])) cost += input->weights[e];
	}
	return cost;
}




void Usage(const char* s)
{
	printf("Usage: %s <input file> <solution file>\n", s);
	exit(1);
}




int main(int argc, char* argv[])
{
	if (argc != 3) Usage(argv[0]);
	TmpGraph<int>* g = ReadDIMACS<int>(argv[1]);
	TmpGraph<int>* T = ReadDIMACS<int>(argv[2]);
	int* x = new int[g->n];

	/*
	int i;
	std::vector<int> V(T->n);
	for (i=0; i<T->n; i++) V[i] = T->weights[i];
	sort( V.begin(),V.end(), [&](int i,int j)
		{return (i<j);}
	);
	for (i=0; i<T->n; i++) std::cout << V[i] << "\n";
	exit(1);
	*/

	
	int e, n = g->n;
	if (n != T->n || T->m != n-1) { printf("sizes don't match\n"); exit(1); }
	for (e=0; e<T->m; e++)
	{
		int s = T->edges[2*e], t = T->edges[2*e+1];

		int cost0 = ComputeCut(g, s, t);
		if (cost0 != T->weights[e])
		{
			std::cout << "(s,t)=(" << s << "," << t << "): true cost=" << cost0 << " cost in GH tree=" << T->weights[e] << "\n";
			std::cout << "ERROR\n"; exit(1);
		}
		ComputeCut(T, e, x);
		int cost1 = 0;
		int e;
		for (e=0; e<g->m; e++)
		{
			if (x[g->edges[2*e]] != x[g->edges[2*e+1]]) cost1 += g->weights[e];
		}
		if (cost0 != cost1)
		{
			std::cout << "(s,t)=(" << s << "," << t << "): true cost=" << cost0 << " cost of GH cut=" << cost1 << "\n";
			std::cout << "ERROR\n"; exit(1);
		}



//		int real_cost = ComputeCut(g, s, t);
//		if (real_cost != T->weights[e])
//		{
//			printf("incorrect Gomory-Hu cut for (s,t)=(%d,%d): true value=%d, stored value=%d\n", s, t, real_cost, T->weights[e]);
//			exit(1);
//		}
	}
	delete [] x;
	printf("cut values for edges in the tree are correct\n");

	return 0;
}








