#define _CRT_SECURE_NO_WARNINGS

#define TMP_GRAPH_IO

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gomory_hu.h"
#include "tmp_graph.h"



void Usage(const char* s)
{
	printf("Usage: %s [options] <filename>\n", s);
	printf("options:\n");
	printf(" -p          use parametric ordered cuts\n");
	printf(" -v<verbose> verbose is non-negative integer\n");
	printf(" -o          save result to <filename>.tree in DIMACS format\n");
	printf(" -O          save result to <filename>.tree in plain format\n");

	exit(1);
}



int main(int argc, char* argv[])
{
	/*
	int argc = 3;
	char* argv[] = { (char*)"a", (char*)"../../test.txt", (char*)"-v3" };
	*/

	int i;
	TmpGraph<GH_REAL>* g = NULL;

	GomoryHuOptions options;
	const char* filename = NULL;
	int saveresult = 0;

	for (i=1; i<argc; i++)
	{
		const char* s = argv[i];
		if (s[0] == '-')
		{
			if      (s[1] == 'p') options.parametric_ordered_cuts = 1;
			else if (s[1] == 'v') options.verbose = atoi(s+2);
			else if (s[1] == 'o') saveresult = 1;
			else if (s[1] == 'O') saveresult = 2;
			else Usage(argv[0]);
		}
		else
		{
			if (filename) Usage(argv[0]);
			filename = s;
		}
	}

	if (!filename) Usage(argv[0]);

	g = ReadDIMACS<GH_REAL>(filename);

	int* parents = new int[g->n];
	GH_REAL* cut_costs = new GH_REAL[g->n];

	SolveGomoryHu(g, parents, cut_costs, options);
	if (saveresult)
	{
		char SAVEFILE[100];
		snprintf(SAVEFILE, sizeof(SAVEFILE), "%s.tree", filename);
		TmpGraph<int> solution(g->n, g->n-1);
		for (i=0; i<g->n; i++)
		{
			if (parents[i] >= 0) solution.AddEdge(i, parents[i], cut_costs[i]);
		}	
		if (saveresult == 1) SaveDIMACS(&solution, SAVEFILE);
		else                 SavePlain(&solution, SAVEFILE);
	}

	delete [] parents;
	delete [] cut_costs;
	delete g;

	return 0;
}

