#ifndef __GOMORY_HU__
#define __GOMORY_HU__

#include "timer.h"
#include "tmp_graph.h"

typedef int GH_REAL;     // GH_REAL should be large enough to hold the cost of any singleton cut in the graph 
typedef long long GH_BIG_REAL; // GH_BIG_REAL should be large enough to hold the cost of an arbitrary cut in the graph

#define REDUCTION_REORDER_EDGES        0x0001
#define REDUCTION_REORDER_NODES        0x0002
#define REDUCTION_CC                   0x0004
#define REDUCTION_MERGE_PARALLEL_EDGES 0x0008

struct GomoryHuOptions
{
	GomoryHuOptions()
		: reduction_method(REDUCTION_REORDER_NODES | REDUCTION_CC),
		  parametric_ordered_cuts(0),
		  verbose(1)
	{
	}
	void Print();
	int reduction_method;
	int parametric_ordered_cuts;
	int verbose;
};


// Note, the call will change 'input', if options.reduction_method != 0
void SolveGomoryHu(TmpGraph<GH_REAL>* input, int* parents, GH_REAL* cut_costs, GomoryHuOptions& options);



// useful util function
inline void ComputePostorder(int n, int* parents, int* postorder)
{
	int i;
	int* child = new int[2*n];
	int* sibling = child + n;

	for (i=0; i<n; i++) child[i] = -1;
	int root = -1;
	for (i=0; i<n; i++)
	{
		int p = parents[i];
		if (p >= 0)
		{
			sibling[i] = child[p];
			child[p] = i;
		}
		else root = i;
	}

	i = root;
	int k = 0;
	while ( 1 )
	{
		if (child[i] >= 0)
		{
			i = child[i];
			continue;
		}
		while ( 1 )
		{
			postorder[k ++] = i;
			if (sibling[i] >= 0) break;
			i = parents[i];
			if (i == root) { postorder[k] = i; break; }
		}
		if (i == root) break;
		i = sibling[i];
	}

	delete [] child;
}

#endif
