#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <string.h>
#include "gomory_hu_reductions.h"

void _SolveGomoryHu(TmpGraph<GH_REAL>* g, int* parents, GH_REAL* cut_costs, GomoryHuOptions& options);

////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

int maxflow_num;
int merge_num;
double maxflow_time;
double maxflow_total_node_num;
double maxflow_total_edge_num;
double GH_total_node_num;
double GH_total_edge_num;

int ComputeTreeDiameter(int n, int* parents)
{
	int k, i;
	int* D = new int[3*n];
	int* E = D + n;
	int* postorder = D + 2*n;
	memset(D, 0, n*sizeof(int));
	ComputePostorder(n, parents, postorder);
	
	int D_max = 0;
	E[postorder[n-1]] = 0;
	for (k=n-2; k>=0; k--)
	{
		i = postorder[k];
		E[i] = 1 + E[parents[i]];
	}
	for (k=0; k<n-1; k++)
	{
		i = postorder[k];
		if (D_max < D[i] + E[i]) D_max = D[i] + E[i];
		int p = parents[i];
		int Dp = D[i] + 1;
		if (D_max < D[p] + Dp) D_max = D[p] + Dp;
		if (D[p] < Dp) D[p] = Dp;
	}
	delete [] D;
	return D_max;
}


void SolveGomoryHu(TmpGraph<GH_REAL> *input, int* parents, GH_REAL* cut_costs, GomoryHuOptions& options)
{
	double T_start = get_time();
	maxflow_num = merge_num = 0;
	maxflow_time = maxflow_total_node_num = maxflow_total_edge_num = GH_total_node_num = GH_total_edge_num = 0;
	int i, n = input->n, m = input->m;

	if (options.verbose >= 1) options.Print();

	GomoryHuReductions R(input, parents, cut_costs, options.reduction_method);
	TmpGraph<GH_REAL>* _input;
	int* _parents;
	GH_REAL* _cut_costs;
	while ((_input=R.GetNextSubproblem(_parents, _cut_costs)))
	{
			_SolveGomoryHu(_input, _parents, _cut_costs, options);
	}

	if (options.verbose >= 1)
	{
		GH_REAL cut_min = 0, cut_max = 0;
		double cut_sum = 0;
		int checksum = 0;
		bool flag = false;
		for (i=0; i<n; i++)
		{
			if (parents[i] < 0) continue;
			if (!flag) { cut_min = cut_max = cut_costs[i]; flag = true; }
			else
			{
				if (cut_min > cut_costs[i]) cut_min = cut_costs[i];
				if (cut_max < cut_costs[i]) cut_max = cut_costs[i];
			}
			cut_sum += cut_costs[i];
			checksum = ((int)(checksum + cut_costs[i])) % 1234;
		}
		std::cout << "cut values: (min,max,average)=(" << cut_min << "," << cut_max << "," << cut_sum/(n-1) << "), checksum=" << checksum << "\n";
		std::cout << "graph size: (n,m) = (" << n << "," << m << "), ";
		std::cout << "tree diameter: " << ComputeTreeDiameter(n, parents) << "\n";
		std::cout << maxflow_num << " maxflows, total graph size = (n*" << ((double)maxflow_total_node_num) / n << ", m*"
			      << ((double)maxflow_total_edge_num) / m << ")\n";
		std::cout << "total GH graph size = (n*" << ((double)GH_total_node_num) / n << ", m*"
			<< ((double)GH_total_edge_num) / m << ")\n";
		std::cout << "maxflow time: " << maxflow_time << " secs\n";
		std::cout << "total   time: " << get_time() - T_start << " secs\n";
	}
}

