#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <vector>
#include <algorithm>
#include <math.h>
#include "ordered_cuts.h"

//#define OC_DEBUG
//#define BK


#ifndef AL_PLUS
#ifndef AL
#ifndef FS_PLUS
#ifndef FS
#define FS_PLUS
#endif
#endif
#endif
#endif

#ifndef BK
#ifndef IBFS
#define IBFS
#endif
#endif

//////////////////////////////////////////////////////////////////
#ifdef AL_PLUS
#define GRAPH_BASE_FILE "graph_base_adjacency_list.h"
#endif
#ifdef AL
#define GRAPH_BASE_FILE "graph_base_adjacency_list.h"
#define KEEP_REV_ARC_POINTER
#endif
#ifdef FS_PLUS
#define GRAPH_BASE_FILE "graph_base_forward_star.h"
#endif
#ifdef FS
#define GRAPH_BASE_FILE "graph_base_forward_star.h"
#define DO_NOT_KEEP_SATURATION_FLAG
#endif

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

#define GRAPH_CONTRACTION

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

#define GRAPH_BASE_REAL GH_REAL

#ifdef BK

#define GRAPH_NAMESPACE ORDERED_CUTS_NEW
#include "bk.inc"
#undef GRAPH_NAMESPACE
typedef ORDERED_CUTS_NEW::Graph G;

#else

#define GRAPH_NAMESPACE ORDERED_CUTS_NEW
#include "ibfs.inc"
#undef GRAPH_NAMESPACE
typedef ORDERED_CUTS_NEW::IBFSGraph G;

#endif


////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////


void GomoryHuOptions::Print()
{
	std::cout << "options: ";

#ifdef BK
	std::cout << " BK algorithm";
#else
	std::cout << " IBFS algorithm";
#endif
#ifdef GRAPH_BASE_FORWARD_STAR
	std::cout << " (forward star";
#ifdef KEEP_SATURATION_FLAG
	std::cout << " +";
#endif
#else
	std::cout << " (adjacency list";
#ifndef KEEP_REV_ARC_POINTER
	std::cout << " +";
#endif
#endif
	if (parametric_ordered_cuts)
		std::cout << ") parametric ordered cuts\n";
	else
		std::cout << ")\n";
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

extern int maxflow_num;
extern double maxflow_time;
extern double maxflow_total_node_num;
extern double maxflow_total_edge_num;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct Component
{
	Component* Split(GH_REAL* _cost)
	{
		if (L < 0) return NULL;

		int i, j, k, e;

		int K;
		if (R < 0)
		{
			K = -R; R *= 2;
			if (K > L/2) K = L/2;
		}
		else
		{
			K = R/2;
		}
		if (K < 1) K = 1;

		G* g = new G(n, m, OC->maxflow_buf, OC->common_rbuf, OC->maxflow_rbuf);

		for (e=0; e<m; e++)
		{
			g->add_edge(edges[2*e], edges[2*e+1], weights[e], weights[e]);
		}
		g->init_graph();
		for (i=0; i<n; i++) UPDATE_EXCESS_FIRST(g, i, unaries[i], 0);

		for (k=0; k<K; k++)
		{
			i = list[k];
			UPDATE_EXCESS_FIRST(g, i, 0, OC->INFTY);
		}

		double T_start = get_time();
		_MAXFLOW_FIRST(g);
		g->set_segment_flags();
		maxflow_time += get_time() - T_start;
		maxflow_num ++;
		maxflow_total_node_num += g->node_num_noncontracted;
		maxflow_total_edge_num += g->edge_num;

		GH_REAL* UNARIES[2];
		GH_REAL* UNARIES_OC[2];
		int* MAPPING_ORIG[2];
		int* index = (int*) OC->common_rbuf->Alloc(n*sizeof(int) + 2*n*sizeof(int) + 4*n*sizeof(GH_REAL));
		MAPPING_ORIG[0] = index + n;
		MAPPING_ORIG[1] = MAPPING_ORIG[0] + n;
		UNARIES[0] = (GH_REAL*) (MAPPING_ORIG[1] + n);
		UNARIES[1] = (GH_REAL*) (UNARIES[0] + n);
		UNARIES_OC[0] = (GH_REAL*) (UNARIES[1] + n);
		UNARIES_OC[1] = (GH_REAL*) (UNARIES_OC[0] + n);
		memset(index, 0, n*sizeof(int));

		for (k=0; k<K; k++) index[list[k]] = 1;
		int N[2] = { 0, 0 };
		t = -1;
		int t_orig = -1;
		if (K == 1 && next == NULL) { t = list[0]; t_orig = mapping_orig[t]; }
		for (i=0; i<n; i++)
		{
			if (i == t) { index[i] = n; continue; }
			int s;
			if (g->nodes[i].segment == 1 || index[i]) { s = 1; index[i] = N[1]; }
			else                                      { s = 0; index[i] = -N[0]-1; }
			MAPPING_ORIG[s][N[s]] = mapping_orig[i];
			UNARIES[s][N[s]]    = unaries[i];
			UNARIES_OC[s][N[s]] = unaries_OC[i];
			N[s] ++;
		}
		OC->maxflow_buf->Reset();
		delete g;

		if (N[0] == 0 && K > 1)
		{
			R = K;
			return NULL;
		}

		Component* Cnext = next;
		next = new Component;
		next->next = Cnext;
		next->OC = OC;

		GH_REAL cost = 0;
		if (t >= 0)
		{
			for (i=0; i<n; i++)
			{
				if (index[i] >= 0) cost += unaries_OC[i];
			}
		}

		int M[2] = { 0, 0 };
		int* EDGES[2];
		GH_REAL* WEIGHTS[2];
		EDGES[0] = (int*) OC->maxflow_buf->Alloc(2*m*sizeof(int));
		EDGES[1] = (int*) OC->maxflow_buf->Alloc(2*m*sizeof(int));
		WEIGHTS[0] = (GH_REAL*) OC->maxflow_buf->Alloc(m*sizeof(GH_REAL));
		WEIGHTS[1] = (GH_REAL*) OC->maxflow_buf->Alloc(m*sizeof(GH_REAL));
		for (e=0; e<m; e++)
		{
			i = index[edges[2*e]];
			j = index[edges[2*e+1]];
			int s;

			GH_REAL w = weights[e];

			if (i == n)
			{
				if (j >= 0 && j < n)
				{
					UNARIES_OC[1][j] += w;
				}
				else if (j < 0)
				{
					UNARIES[0][-1-j] -= w;
					UNARIES_OC[0][-1-j] += w;
					cost += w;
				}
				continue;
			}
			if (j == n)
			{
				if (i >= 0 && i < n)
				{
					UNARIES_OC[1][i] += w;
				}
				else if (i < 0)
				{
					UNARIES[0][-1-i] -= w;
					UNARIES_OC[0][-1-i] += w;
					cost += w;
				}
				continue;
			}

			if (i >= 0 && j >= 0) s = 1;
			else if (i < 0 && j < 0) { i = -1-i; j = -1-j; s = 0; }
			else 
			{
				if (i < 0)
				{
					UNARIES[0][-1-i] -= w;
					UNARIES[1][j] += w;
					UNARIES_OC[0][-1-i] += w;
					UNARIES_OC[1][j] += w;
				}
				else
				{
					UNARIES[0][-1-j] -= w;
					UNARIES[1][i] += w;
					UNARIES_OC[0][-1-j] += w;
					UNARIES_OC[1][i] += w;
				}
				cost += w;
				continue;
			}

			EDGES[s][2*M[s]] = i;
			EDGES[s][2*M[s]+1] = j;
			WEIGHTS[s][M[s] ++] = weights[e];
		}

		int n_old = n;
		n = N[0];
		next->n = N[1];
		m = M[0];
		next->m = M[1];
		next->edges = edges + 2*m;
		next->weights = weights + m;
		next->unaries = unaries + n;
		next->unaries_OC = unaries_OC + n;
		memcpy(edges, EDGES[0], 2*m*sizeof(int));
		memcpy(weights, WEIGHTS[0], m*sizeof(GH_REAL));
		memcpy(next->edges, EDGES[1], 2*next->m*sizeof(int));
		memcpy(next->weights, WEIGHTS[1], next->m*sizeof(GH_REAL));
		memcpy(unaries, UNARIES[0], n*sizeof(GH_REAL));
		memcpy(next->unaries, UNARIES[1], next->n*sizeof(GH_REAL));
		memcpy(unaries_OC, UNARIES_OC[0], n*sizeof(GH_REAL));
		memcpy(next->unaries_OC, UNARIES_OC[1], next->n*sizeof(GH_REAL));

		OC->maxflow_buf->Reset();

		next->mapping_orig = mapping_orig + n;
		memcpy(mapping_orig, MAPPING_ORIG[0], n*sizeof(int));
		memcpy(next->mapping_orig, MAPPING_ORIG[1], next->n*sizeof(int));

		int LL[2] = { 0, 0 };
		int* LIST[2];
		LIST[0] = (int*) OC->maxflow_buf->Alloc(N[0]*sizeof(int));
		LIST[1] = (int*) OC->maxflow_buf->Alloc(N[1]*sizeof(int));

		int R_old = R;
		for (k=0; k<L; k++)
		{
			i = list[k];
			if (i != t)
			{
				i = index[i];
				if (i >= 0)
				{
					LIST[1][LL[1] ++] = i;
				}
				else
				{
					LIST[0][LL[0] ++] = -1-i;
				}
			}
			if (k == K-1) next->R = LL[1];
			if (k == R_old-1) R = LL[0];
		}
		next->t = t_orig;
		t = -1;
		L = LL[0];
		next->L = LL[1];
		next->list = list + L;
		memcpy(list, LIST[0], L*sizeof(int));
		memcpy(next->list, LIST[1], next->L*sizeof(int));

		OC->maxflow_buf->Reset();

		if (t_orig < 0) return NULL;
		*_cost = cost;
		return next;
	}

	int n, m;
	int* edges;
	GH_REAL* weights;
	GH_REAL* unaries;
	GH_REAL* unaries_OC;
	int* mapping_orig;
	int L; 
	int R; // if > 0: // list[0],...,list[R-1] gives the full graph cut
	       // if < 0: // let's try list[0],...,list[-R]
	int* list;
	int t;
	Component* next;

	OrderedCuts* OC;
};


////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

OrderedCuts::OrderedCuts(GH_REAL _INFTY, GomoryHuOptions* _options, ReusableBuffer* _common_rbuf)
	: INFTY(_INFTY), options(_options), depth(0), common_rbuf(_common_rbuf), parent(NULL), child(NULL)
{
	maxflow_buf = new Buffer(1024);
	maxflow_rbuf = new ReusableBuffer[MAXFLOW_RBUF_NUM+1];
}

OrderedCuts::OrderedCuts(OrderedCuts* _parent)
	: INFTY(_parent->INFTY), options(_parent->options), depth(_parent->depth+1), common_rbuf(_parent->common_rbuf), child(NULL), parent(_parent)
{
	maxflow_buf = parent->maxflow_buf;
	maxflow_rbuf = parent->maxflow_rbuf;
}

OrderedCuts::~OrderedCuts()
{
	if (child) delete child;
	if (!parent)
	{
		delete maxflow_buf;
		delete [] maxflow_rbuf;
	}
}

void OrderedCuts::SolveParametric(TmpGraph<GH_REAL>* input, GH_REAL* unaries, int L, int* list, Node* nodes_orig, int s_weight, GH_REAL* REAL_buf)
{
	int i, n = input->n, m = input->m;

	int* mapping_orig = new int[n];
	for (i=0; i<n; i++) mapping_orig[i] = i;
	for (i=0; i<n; i++) { nodes_orig[i].parent = -1; nodes_orig[i].cost = 0; }
	GH_REAL* unaries_OC = (REAL_buf) ? REAL_buf : (new GH_REAL[n]);
	memcpy(unaries_OC, unaries, n*sizeof(GH_REAL));

	Component* C = new Component;
	C->n = n;
	C->m = m;
	C->edges = input->edges;
	C->weights = input->weights;
	C->unaries = unaries;
	C->unaries_OC = unaries_OC;
	C->L = L;
	C->R = -s_weight;
	C->t = -1;
	C->list = list;
	C->mapping_orig = mapping_orig;
	C->next = NULL;
	C->OC = this;

	s_weight *= 2;

	while (C)
	{
		if (C->L <= 0)
		{
			Component* Ctmp = C;
			C = C->next;
			delete Ctmp;
			continue;
		}
		if (C->R == 1 || C->t >= 0)
		{
			TmpGraph<GH_REAL> I;
			I.n = C->n;
			I.m = C->m;
			I.edges = C->edges;
			I.weights = C->weights;

			Node* Cnodes = new Node[C->n];
			SolveParametric(&I, C->unaries_OC, C->L, C->list, Cnodes, s_weight, C->unaries);
			for (i=0; i<C->n; i++)
			{
				if (Cnodes[i].parent >= 0)
				{
					nodes_orig[C->mapping_orig[i]].parent = C->mapping_orig[Cnodes[i].parent];
				}
				nodes_orig[C->mapping_orig[i]].cost   = Cnodes[i].cost;
			}
			delete [] Cnodes;

			I.edges = NULL;
			I.weights = NULL;

			Component* Ctmp = C;
			C = C->next;
			delete Ctmp;
		}
		else
		{
			GH_REAL cost;
			Component* C0 = C;
			Component* Cnext0 = C->next;
			Component* Ct = C->Split(&cost);
			if (Ct)
			{
				for (i=0; i<Ct->n; i++)
				{
					nodes_orig[Ct->mapping_orig[i]].parent = Ct->t;
					nodes_orig[Ct->mapping_orig[i]].cost = cost;
				}
				nodes_orig[Ct->t].cost = cost;
			}
		}
	}

	delete [] mapping_orig;
	if (!REAL_buf) delete [] unaries_OC;
}

void OrderedCuts::Solve(TmpGraph<GH_REAL>* input, GH_REAL* unaries, int L, int* list, Node* nodes_orig, int s_weight)
{
	int i, j, k, e, K;
#ifdef OC_DEBUG
	for (i=0; i<depth; i++) printf(" ");
	printf("L=%d,s=%d  %d=>", L, s_weight, input->n);
#endif

	int n_orig = input->n;
	int* mapping_orig = (int*) local_rbuf1.Alloc(2*n_orig*sizeof(int));
	for (i=0; i<n_orig; i++) mapping_orig[i] = i;
	int* index = mapping_orig + n_orig;

	while ( L > 0 )
	{
		int n = input->n;

		K = s_weight;
		if (K > L/2) K = L/2;
		if (K < 1) K = 1;
	
		G* g = new G(n, input->m, maxflow_buf, common_rbuf, maxflow_rbuf);

		//g->init_graph(input);
		for (e=0; e<input->m; e++)
		{
			g->add_edge(input->edges[2*e], input->edges[2*e+1], input->weights[e], input->weights[e]);
		}
		g->init_graph();
		for (i=0; i<n; i++) UPDATE_EXCESS_FIRST(g, i, unaries[i], 0);

		for (k=0; k<K; k++)
		{
			i = list[k];
			UPDATE_EXCESS_FIRST(g, i, 0, INFTY);
		}

		double T_start = get_time();
		_MAXFLOW_FIRST(g);
		g->set_segment_flags();
		maxflow_time += get_time() - T_start;
		maxflow_num ++;
		maxflow_total_node_num += g->node_num_noncontracted;
		maxflow_total_edge_num += g->edge_num;

		int n1=0, n2=0;
		for (i=0; i<n; i++)
		{
			if (g->nodes[i].segment == 1) { index[i] = 1; n2 ++; }
			else index[i] = 0;
		}
		maxflow_buf->Reset();
		delete g;

		for (k=0; k<K; k++) { if (index[list[k]] > 0) continue; index[list[k]] = 1; n2 ++; }
		int t = (K == 1) ? list[0] : -1;
		if (t >= 0) index[t] = n;

		// n2 = 0    =>    s_weight = 2*K
		// n2 = n    =>    s_weight = 0.5*K
		double alpha = (n - 2.0*n2) / n;
		int s_weight2 = (int)(K*pow(2,alpha) + 0.5);
		if (n2 == n) s_weight = K/2;
		s_weight = (int)(K*pow(2,alpha) + 0.5);


		int* index2_rev = (int*) local_rbuf2.Alloc(n2*(sizeof(int) + sizeof(Node)));
		Node* nodes2 = (Node*)(index2_rev + n2);

		n1 = n2 = 0;
		for (i=0; i<n; i++)
		{
			if (index[i] > 0) { if (i != t) { index2_rev[n2] = i; index[i] = n2 ++; } }
			else               index[i] = -1 - (n1++);
		}
#ifdef OC_DEBUG
		printf("%d,%d\n", n1, n2);
#endif

		char* buf_local = (char*) common_rbuf->Alloc(2*input->m*sizeof(int) + input->m*sizeof(GH_REAL) + n2*sizeof(GH_REAL) + n2*sizeof(int));
		int* edges2 = (int*) buf_local;
		GH_REAL* weights2 = (GH_REAL*) (edges2 + 2*input->m);
		GH_REAL* unaries2 = weights2 + input->m;
		int* list2 = (int*) (unaries2 + n2);

		GH_REAL cost = 0;
		for (i=0; i<n; i++)
		{
			k = index[i];
			if (k >= 0)
			{
				if (k != n) unaries2[k] = unaries[i];
				cost += unaries[i];
			}
			else
			{
				unaries[-1-k] = unaries[i];
			}
		}
		int m1 = 0, m2 = 0;
		for (e=0; e<input->m; e++)
		{
			i = index[input->edges[2*e]];
			j = index[input->edges[2*e+1]];
			GH_REAL w = input->weights[e];
			if (i >= 0)
			{
				if (j >= 0)
				{
					if      (i == n) unaries2[j] += w;
					else if (j == n) unaries2[i] += w;
					else { edges2[2*m2] = i; edges2[2*m2+1] = j; weights2[m2++] = w; }
				}
				else
				{
					unaries[-1-j] += w; cost += w;
					if (i != n) unaries2[i] += w;
				}
			}
			else
			{
				if (j >= 0)
				{
					unaries[-1-i] += w; cost += w;
					if (j != n) unaries2[j] += w;
				}
				else { input->edges[2*m1] = -1-i; input->edges[2*m1+1] = -1-j; input->weights[m1++] = w; }
			}
		}
		int L1 = 0, L2 = 0;
		for (k=0; k<L; k++)
		{
			i = index[list[k]];
			if (i >= 0) { if (i != n) list2[L2 ++] = i; }
			else        list[L1 ++] = -1-i;
		}

		input->n = n1;
		input->m = m1;
		L = L1;
		memcpy(input->edges + 2*m1, edges2, 2*m2*sizeof(int)); edges2 = input->edges + 2*m1;
		memcpy(input->weights + m1, weights2, m2*sizeof(GH_REAL)); weights2 = input->weights + m1;
		memcpy(unaries + n1, unaries2, n2*sizeof(GH_REAL)); unaries2 = unaries + n1;
		memcpy(list + L1, list2, L2*sizeof(int)); list2 = list + L1;
		//free(buf_local);

		TmpGraph<GH_REAL> input2;
		input2.n = n2;
		input2.m = m2;
		input2.edges = edges2;
		input2.weights = weights2;

		if (!child) child = new OrderedCuts(this);

		child->Solve(&input2, unaries2, L2, list2, nodes2, s_weight2); // + ((K==1) ? 1 : 0));
		input2.edges = NULL;
		if (t >= 0)
		{
			t = mapping_orig[t];
			nodes_orig[t].parent = -1;
			nodes_orig[t].cost = cost;
		}
		for (i=0; i<n2; i++)
		{
			if (nodes2[i].parent >= 0)
			{
				nodes_orig[mapping_orig[index2_rev[i]]].parent = mapping_orig[index2_rev[nodes2[i].parent]];
			}
			else
			{
				nodes_orig[mapping_orig[index2_rev[i]]].parent = t;
			}
			nodes_orig[mapping_orig[index2_rev[i]]].cost = nodes2[i].cost;
		}
		for (i=0; i<n; i++)
		{
			k = index[i];
			if (k < 0) mapping_orig[-1-k] = mapping_orig[i];
		}
	}

	GH_REAL cost = 0;
	for (i=0; i<input->n; i++) cost += unaries[i];
	for (i=0; i<input->n; i++)
	{
		nodes_orig[mapping_orig[i]].parent = -1;
		nodes_orig[mapping_orig[i]].cost = cost;
	}
}

