#ifndef __TMP_GRAPH__
#define __TMP_GRAPH__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "block.h"

template <class Weight> class TmpGraph
{
public:
	TmpGraph() : edges(NULL), permutation(NULL), permutation_rev(NULL), marks(NULL) {}
	TmpGraph(int _n, int _m_max) : n(_n), m_max(_m_max), m(0), permutation(NULL), permutation_rev(NULL), marks(NULL)
	{
		edges = (int*) malloc(2*m_max*sizeof(int) + m_max*sizeof(Weight));
		weights = (Weight*) (edges + 2*m_max);
		component_num = 0;
	}
	TmpGraph(TmpGraph<Weight> *g) : n(g->n), m(g->m), m_max(g->m), permutation(NULL), permutation_rev(NULL), marks(NULL)
	{
		edges = (int*) malloc(2*m*sizeof(int) + m*sizeof(Weight));
		weights = (Weight*) (edges + 2*m);
		memcpy(edges, g->edges, 2*m*sizeof(int) + m*sizeof(Weight));
		if (g->marks)
		{
			marks = new int[n];
			memcpy(marks, g->marks, n*sizeof(int));
		}
		component_num = 1;
		component_sizes = (int*) rbuf.Realloc(sizeof(int));
		component_sizes[0] = n;
		component_num = 0;
	}
	~TmpGraph()
	{
		if (edges) free(edges);
		if (permutation) delete [] permutation;
		if (marks) delete [] marks;
	}

	void AddEdge(int i, int j, Weight w)
	{
		if (m >= m_max)
		{
			printf("Too many edges!\n"); exit(1);
		}
		edges[2*m] = i; edges[2*m+1] = j; weights[m++] = w;
	}

	void UnsetMark(int i)
	{
		if (!marks) { marks = new int[n]; int j; for (j=0; j<n; j++) marks[j] = 1; }
		marks[i] = 0;
	}
	bool isMarked(int i) { if (marks) return (bool)(marks[i]); return true; }

	int Check() // eliminate self-loops, add constant to node ids to make the minimum id zero
	{
		int e;
		if (m == 0) return 0;
		int i_min = edges[0], i_max = edges[0];
		for (e=1; e<2*m; e++)
		{
			if (i_min > edges[e]) i_min = edges[e];
			if (i_max < edges[e]) i_max = edges[e];
		}
		i_max -= i_min;
		if (i_max >= n) { printf("Error: wrong index\n"); exit(1); }
		for (e=0; e<2*m; e++) edges[e] -= i_min;
		for (e=0; e<m; e++)
		{
			if (edges[2*e] == edges[2*e+1])
			{
				m --;
				edges[2*e] = edges[2*m];
				edges[2*e+1] = edges[2*m+1];
				weights[e] = weights[m];
				e --;
			}
			if (weights[e] < 0) { printf("Error: negative weight\n"); exit(1); }
		}
		return i_min;
	}

	void ReorderNodesAndEdges()
	{
		if (!permutation)
		{
			permutation = new int[2*n]; permutation_rev = permutation + n;
		}

		int i, j, i0, N, e, e0, ee;

 		int* degrees = (int*) malloc((2*n+2*m)*sizeof(int) + m*sizeof(Weight) + 2*m*sizeof(char));
		int* degrees0 = degrees + n;
		int* edges_tmp = degrees0 + n;
		Weight* weights_tmp = (Weight*)(edges_tmp + 2*m);
		char* flags = (char*)(weights_tmp + m);

		memset(degrees, 0, 2*n*sizeof(int));
		for (e=0; e<m; e++)
		{
			i = edges[2*e]; j = edges[2*e+1];
			degrees0[i] ++;
			degrees[i] ++;
			degrees[j] ++;
		}
		for (i=e=e0=0; i<n; i++)
		{
			int d = degrees0[i];
			degrees0[i] = e0;
			e0 += d;
			d = degrees[i];
			degrees[i] = e;
			e += d;
		}
		for (e=0; e<m; e++)
		{
			i = edges[2*e]; j = edges[2*e+1];
			weights_tmp[degrees0[i] ++] = weights[e];
			flags[degrees[i]] = 1;
			flags[degrees[j]] = 0;
			edges_tmp[degrees[i] ++] = j;
			edges_tmp[degrees[j] ++] = i;
		}

		component_num = 0;
		int N_prev = 0;

		for (i=0; i<n; i++) permutation_rev[i] = -1;
		for (i0=N=0; i0<n; i0++)
		{
			if (permutation_rev[i0] >= 0) continue;
			int A = N, Mc = 0;
			permutation[N] = i0; permutation_rev[i0] = N ++;
			while (A < N)
			{
				i = permutation[A ++];
				for (e=(i==0) ? 0 : degrees[i-1]; e<degrees[i]; e++, Mc ++)
				{
					j = edges_tmp[e];
					if (permutation_rev[j] >= 0) continue;
					permutation[N] = j; permutation_rev[j] = N ++;
				}
			}
			component_sizes = (int*) rbuf.Realloc((2*(component_num + 1))*sizeof(int));
			component_sizes[2*component_num] = N - N_prev;
			component_sizes[2*component_num+1] = Mc / 2;
			component_num ++;
			N_prev = N;
		}
		for (i0=ee=0; i0<n; i0++)
		{
			i = permutation[i0];
			if (i == 0) e = e0 = 0;
			else { e = degrees[i-1]; e0 = degrees0[i-1]; }
			for ( ; e<degrees[i]; e++)
			{
				if (flags[e])
				{
					edges[2*ee] = i0;
					edges[2*ee+1] = permutation_rev[edges_tmp[e]];
					weights[ee ++] = weights_tmp[e0 ++];
				}
			}
		}

		if (marks)
		{
			memcpy(degrees, marks, n*sizeof(int));
			for (i=0; i<n; i++) marks[i] = degrees[permutation[i]];
		}

		free(degrees);
	}

	void ReorderEdges(bool merge_parallel_edges=false) // reorder edges to get better locality
	{
		int* degree = new int[n];
		memset(degree, 0, n*sizeof(int));
		int i, e;
		for (e=0; e<m; e++)
		{
			i = edges[2*e];
			if (merge_parallel_edges && i > edges[2*e+1]) { edges[2*e] = edges[2*e+1]; edges[2*e+1] = i; i = edges[2*e]; }
			degree[i] ++;
		}
		int* edges_new = (int*) malloc(2*m*sizeof(int) + m*sizeof(Weight));
		Weight* weights_new = (Weight*) (edges_new + 2*m);
		e = 0;
		for (i=0; i<n; i++)
		{
			int d = degree[i];
			degree[i] = e;
			e += d;
		}
		for (e=0; e<m; e++)
		{
			i = edges[2*e];
			int d = degree[i] ++;
			edges_new[2*d] = edges[2*e];
			edges_new[2*d+1] = edges[2*e+1];
			weights_new[d] = weights[e];
		}
		if (merge_parallel_edges)
		{
			int* flags = edges;
			bool allocate_flags = false;
			if (n > 2*m) { allocate_flags = true; flags = new int[n]; }
			for (i=0; i<n; i++) flags[i] = -1;
			int from = 0, e_new = 0;
			for (i=0; i<n; i++)
			{
				int to = degree[i];
				int from_new = e_new;
				for (e=from; e<to; e++)
				{
					int j = edges_new[2*e+1];
					if (flags[j] < 0)
					{
						edges_new[2*e_new] = edges_new[2*e];
						edges_new[2*e_new+1] = edges_new[2*e+1];
						weights_new[e_new] = weights_new[e];
						flags[j] = e_new ++;
					}
					else
					{
						weights_new[flags[j]] += weights_new[e];
					}
				}
				for (e=from_new; e<e_new; e++)
				{
					flags[edges_new[2*e+1]] = -1;
				}
				from = to;
			}
			m = e_new;
			if (allocate_flags) delete [] flags;
		}
		free(edges);
		edges = edges_new;
		weights = weights_new;
		delete [] degree;
	}

	void Print()
	{
		int e;
		for (e=0; e<m; e++) printf("%d %d %d\n", edges[2*e], edges[2*e+1], weights[e]);
	}

	int n, m, m_max;
	int* edges; // of size 2*m
	Weight* weights; // of size m

	int* marks; // of size n, if not NULL

	// these variables are set by ReorderNodesAndEdges()
	int* permutation;
	int* permutation_rev;
	int component_num;
	int* component_sizes; // of size 2*component_num. component_size[2*C] = # nodes, component_size[2*C+1] = # edges
	ReusableBuffer rbuf; // for 'components'
};

#ifdef TMP_GRAPH_IO

template <typename REAL>
inline TmpGraph<REAL>* ReadPlain(const char* filename)
{
	FILE* fp = fopen(filename, "r");
	if (!fp) { printf("Cannot open %s\n", filename); exit(1); }

	int n, m=0, m_max = 16, i, j, w;

	int* edges = (int*) malloc(2*m_max*sizeof(int));
	int* weights = (int*) malloc(m_max*sizeof(int));

	char LINE[1000];
	int self_loop_num = 0;
	while (fgets(LINE, sizeof(LINE), fp))
	{
		bool success = false;
		if (sscanf(LINE, "%d %d %d", &i, &j, &w) == 3) success = true;
		else if (sscanf(LINE, "%d %d", &i, &j) == 2) { w = 1; success = true; }
		if (success)
		{
			if (m >= m_max) { m_max *= 2; edges = (int*) realloc(edges, 2*m_max*sizeof(int)); weights = (int*) realloc(weights, m_max*sizeof(int)); }
			if (i == j) { self_loop_num ++; continue; }
			edges[2*m] = i; edges[2*m+1] = j; weights[m++] = w;
		}
	}
	if (self_loop_num > 0) printf("%d self-loops (ignored)\n", self_loop_num); 

	if (m < 1) { printf("%s: no edges?\n", filename); exit(1); }
	int i_min = edges[0], i_max = edges[0];
	for (i=0; i<2*m; i++)
	{
		if (i_min > edges[i]) i_min = edges[i];
		if (i_max < edges[i]) i_max = edges[i];
	}
	n = i_max - i_min + 1;
	printf("%s: %d nodes, %d edges\n", filename, n, m);
	TmpGraph<REAL>* g = new TmpGraph<REAL>(n,m);
	for (i=0; i<m; i++) g->AddEdge(edges[2*i]-i_min, edges[2*i+1]-i_min, weights[i]);
	free (edges);
	free (weights);

	fclose(fp);
	return g;
}

template <typename REAL>
inline TmpGraph<REAL>* ReadDIMACS(const char* filename, int* s=NULL, int* t=NULL)
{
	if (s) *s = -1;
	if (t) *t = -1;
	FILE* fp = fopen(filename, "r");
	if (!fp) { printf("Cannot open %s\n", filename); exit(1); }

	TmpGraph<REAL>* g = NULL;
	int n, m, i, j, w;
	char c;

	char LINE[1000];
	int self_loop_num = 0;
	while (fgets(LINE, sizeof(LINE), fp))
	{
		switch (LINE[0])
		{
			case 'p':
				if (g) { printf("Error in %s: multiple 'p' lines\n", filename); exit(1); }
				if (sscanf(LINE, "%*c %*3s %d %d", &n, &m) != 2 || n < 1 || m < 0)
				{
					printf("Error in %s: wrong 'p' line: %s\n", filename, LINE); exit(1);
				}
				g = new TmpGraph<REAL>(n, m);
				break;
			case 'a':
				if (!g) { printf("Error in %s: no 'p' line before 'a' line\n", filename); exit(1); }
				if (sscanf(LINE, "a %d %d %d", &i, &j, &w) != 3) { printf("Error in %s: wrong 'a' line: %s\n", filename, LINE); exit(1); }
				if (i == j) self_loop_num ++;
				else g->AddEdge(i, j, w);
				break;
			case 'n':
				if (!g) { printf("Error in %s: no 'p' line before 'n' line\n", filename); exit(1); }
				if (sscanf(LINE, "n %d %c", &i, &c) == 2)
				{
					if (c == 's') { if (!s)  { printf("Error in %s: line %s\n", filename, LINE); exit(1); } *s = i; break; }
					if (c == 't') { if (!t)  { printf("Error in %s: line %s\n", filename, LINE); exit(1); } *t = i; break; } 
				}
				printf("Error in %s: wrong 'n' line: %s\n", filename, LINE); exit(1);
			default:
				break;
		}
	}
	fclose(fp);
	if (self_loop_num > 0) printf("%d self-loops (ignored)\n", self_loop_num); 
	if (!g)
	{
		printf("trying plain format...\n");
		return ReadPlain<REAL>(filename);
//		printf("Error in %s: no 'p' line\n", filename); exit(1);
	}
	if (g->m != g->m_max) { printf("Error in %s: too few edges (%d < %d)\n", filename, g->m, g->m_max); exit(1); }


	int i_min = g->Check();
	if (s) *s -= i_min;
	if (t) *t -= i_min;
	return g;
}



inline void SaveDIMACS(TmpGraph<int>* g, const char* filename)
{
	FILE* fp = fopen(filename, "w");
	if (!fp) { printf("Cannot open %s for writing\n", filename); exit(1); }
	fprintf(fp, "p cut %d %d\n", g->n, g->m);
	int e;
	for (e=0; e<g->m; e++) fprintf(fp, "a %d %d %d\n", g->edges[2*e]+1, g->edges[2*e+1]+1, g->weights[e]);
	fclose(fp);
}

inline void SavePlain(TmpGraph<int>* g, const char* filename)
{
	FILE* fp = fopen(filename, "w");
	if (!fp) { printf("Cannot open %s for writing\n", filename); exit(1); }
	int e;
	bool unit_weights = true;
	for (e=0; e<g->m; e++) if (g->weights[e] != 1) unit_weights = false;
	if (unit_weights)
	{
		for (e=0; e<g->m; e++) fprintf(fp, "%d %d\n", g->edges[2*e], g->edges[2*e+1]);
	}
	else
	{
		for (e=0; e<g->m; e++) fprintf(fp, "%d %d %d\n", g->edges[2*e], g->edges[2*e+1], g->weights[e]);
	}
	fclose(fp);
}

#endif

#endif
